"""Provides the PropertyTypes class."""

from datetime import datetime, timedelta

NoneType = type(None)


property_types = dict(
    null=NoneType,
    bool=bool,
    str=str,
    float=float,
    int=int,
    list=list,
    dict=dict,
    datetime=datetime,
    duration=timedelta
)



"""Test the events framework."""

from attr import attrs, attrib
from pytest import raises
from carehome import Database, methods
from carehome.exc import NoSuchEventError

db = Database()


@attrs
class Stuff:
    args = attrib()
    kwargs = attrib()


stuff = Stuff(None, None)
methods.stuff = stuff
code = '''def %s(self, *args, **kwargs):
    stuff.args = args
    stuff.kwargs = kwargs
'''


def test_valid_event():
    o = db.create_object()
    o.add_method(
        'def on_event(self, *args, **kwargs):\n    return (self, args, kwargs)'
    )
    self, args, kwargs = o.do_event('on_event', 1, 2, 3, hello='world')
    assert self is o
    assert args == (1, 2, 3)
    assert kwargs == {'hello': 'world'}


def test_invalid_event():
    o = db.create_object()
    with raises(NoSuchEventError):
        o.do_event('test_event')


def test_try_event_valid():
    o = db.create_object()
    o.add_method(
        'def on_event(self, *args, **kwargs):\n    return (self, args, kwargs)'
    )
    self, args, kwargs = o.try_event('on_event', 1, 2, 3, hello='world')
    assert self is o
    assert args == (1, 2, 3)
    assert kwargs == {'hello': 'world'}


def test_try_event_invalid():
    o = db.create_object()
    assert o.try_event('test_event') is None


def test_on_init():
    p = db.create_object()
    p.add_method(code % 'on_init')
    o = db.create_object(p)
    assert len(stuff.args) == 1
    assert stuff.args[0] is o
    assert stuff.kwargs == {}


def test_on_attach():
    o = db.create_object()
    o.add_method(code % 'on_attach')
    db.attach_object(o)
    assert len(stuff.args) == 1
    assert stuff.args[0]is o
    assert stuff.kwargs == {}


def test_on_destroy():
    o = db.create_object()
    o.add_method(code % 'on_destroy')
    db.destroy_object(o)
    assert len(stuff.args) == 1
    assert stuff.args[0] is o
    assert stuff.kwargs == {}


def test_on_exit():
    loc = db.create_object()
    loc.add_method(
        'def on_exit(self, obj, thing):\n    self.last_left = thing'
    )
    thing = db.create_object()
    thing.location = loc
    assert not loc.properties
    thing.location = None
    assert loc.last_left is thing
    other = db.create_object()
    other.location = loc
    assert loc.last_left is thing
    other.location = None
    assert loc.last_left is other


def test_on_enter():
    loc = db.create_object()
    loc.add_method(
        'def on_enter(self, obj, thing):\n    self.last_entered = thing'
    )
    assert not loc.properties
    thing = db.create_object()
    thing.location = loc
    assert loc.last_entered is thing
    assert loc.contents == [thing]
    other = db.create_object()
    other.location = loc
    assert loc.last_entered is other
    assert loc.contents == [thing, other]


def test_on_add_parent():
    p = db.create_object()
    c = db.create_object()
    c.add_method(code % 'on_add_parent')
    c.add_parent(p)
    assert len(stuff.args) == 2
    assert stuff.args[0] is c
    assert stuff.args[1] is p
    assert stuff.kwargs == {}


def test_on_add_child():
    p = db.create_object()
    c = db.create_object()
    p.add_method(code % 'on_add_child')
    c.add_parent(p)
    assert len(stuff.args) == 2
    assert stuff.args[0] is p
    assert stuff.args[1] is c
    assert stuff.kwargs == {}


def test_on_remove_parent():
    p = db.create_object()
    c = db.create_object(p)
    c.add_method(code % 'on_remove_parent')
    c.remove_parent(p)
    assert len(stuff.args) == 2
    assert stuff.args[0] is c
    assert stuff.args[1] is p
    assert stuff.kwargs == {}


def test_on_remove_child():
    p = db.create_object()
    p.add_method(code % 'on_remove_child')
    c = db.create_object(p)
    c.remove_parent(p)
    assert len(stuff.args) == 2
    assert stuff.args[0] is p
    assert stuff.args[1] is c
    assert stuff.kwargs == {}


def test_on_add_property():
    o = db.create_object()
    o.add_method(code % 'on_add_property')
    p = o.add_property('test', str, 'Testing.')
    assert len(stuff.args) == 2
    assert stuff.args[0] is o
    assert stuff.args[1] is p
    assert stuff.kwargs == {}


def test_on_remove_property():
    o = db.create_object()
    o.add_method(code % 'on_remove_property')
    name = 'test'
    o.add_property(name, str, 'Testing.')
    o.remove_property(name)
    assert len(stuff.args) == 2
    assert stuff.args[0] is o
    assert stuff.args[1] == name
    assert stuff.kwargs == {}



"""Clear the methods directory before we start."""

from shutil import rmtree
from pytest import fixture


@fixture(scope='session', autouse=True)
def clear_method_files():
    """First we yield to start the tests."""
    yield  # Test, test, test!
    rmtree('methods')



"""Test Property objects."""

from datetime import datetime
from pytest import raises
from carehome import Property

name = 'test'


def test_init():
    description = 'Test property.'
    value = datetime.utcnow()
    p = Property(name, description, datetime, value)
    assert p.name == name
    assert p.description == description
    assert p.type is datetime
    assert p.value == value


def test_get():
    value = 'Test'
    p = Property(name, 'Test property.', str, value)
    assert p.get() is value
    value = 'Some other value.'
    p.value = value
    assert p.get() is value


def test_set():
    value = 'Test value'
    p = Property(name, 'Test property.', str, value)
    with raises(TypeError):
        p.set(datetime.utcnow())
    assert p.value is value
    value = 'Some other value.'
    p.set(value)
    assert p.value is value
    p.set(None)
    assert p.value is None



"""Provides the Database class."""

import os
import os.path
from attr import attrs, attrib, Factory
from .exc import (
    LoadPropertyError, LoadMethodError, LoadObjectError, ObjectRegisteredError,
    HasChildrenError, HasContentsError, IsValueError
)
from .objects import Object
from .properties import Property
from .methods import Method
from .property_types import property_types


@attrs
class ObjectReference:
    """A reference to an object. used when dumping and loading properties."""

    id = attrib()


@attrs
class Database:
    """A database which holds references to objects, methods to create and
    destroy them, and the current max ID."""

    objects = attrib(default=Factory(dict), init=False, repr=False)
    max_id = attrib(default=Factory(int), init=False)
    registered_objects = attrib(default=Factory(dict), init=False, repr=False)
    object_class = attrib(default=Factory(lambda: Object))
    property_class = attrib(default=Factory(lambda: Property))
    method_class = attrib(default=Factory(lambda: Method))
    property_types = attrib(default=Factory(lambda: property_types.copy()))
    method_globals = attrib(default=Factory(type(None)))
    methods_dir = attrib(default=Factory(lambda: 'methods'))

    def __attrs_post_init__(self):
        if not os.path.isdir(self.methods_dir):
            os.makedirs(self.methods_dir)
        if self.method_globals is None:
            self.method_globals = dict(database=self)
        self.method_globals.setdefault('objects', self.objects)
        self.property_types['obj'] = self.object_class

    def new_id(self):
        """Get a unique ID and increment self.max_id."""
        self.max_id += 1
        return self.max_id - 1

    def create_object(self, *parents):
        """Create an object that will be added to the dictionary of objects.
        This object will have all the provided parents added to it."""
        o = self.object_class(self, id=self.new_id())
        for parent in parents:
            o.add_parent(parent)
        self.attach_object(o)
        o.try_event('on_init', o)
        return o

    def attach_object(self, o):
        """Attach an Object instance o to this database."""
        self.max_id = max(o.id + 1, self.max_id)
        self.objects[o.id] = o
        o.try_event('on_attach', o)

    def test_value(self, value, obj):
        """Return True if obj is found somewhere in value."""
        if value is obj:
            return True
        elif isinstance(value, list):
            for entry in value:
                if self.test_value(entry, obj):
                    return True
            else:
                return False
        elif isinstance(value, dict):
            for name, data in value.items():
                if self.test_value(data, obj):
                    return True
            else:
                return False
        else:
            return False

    def destroy_object(self, obj):
        """Destroy an object obj."""
        for name, value in self.registered_objects.items():
            if value is obj:
                raise ObjectRegisteredError(name, value)
        if obj.children:
            raise HasChildrenError(obj)
        if obj.contents:
            raise HasContentsError(obj)
        obj.try_event('on_destroy', obj)
        for thing in self.objects.values():
            for prop in thing._properties.values():
                if self.test_value(prop.value, obj):
                    raise IsValueError(thing, prop)
        for parent in obj._parents:
            obj.remove_parent(parent)
        del self.objects[obj.id]

    def dump_value(self, value):
        """Return a properly dumped value. Used for converting Object instances
        to ObjectReference instances."""
        if isinstance(value, self.object_class):
            return ObjectReference(value.id)
        elif isinstance(value, list):
            return [self.dump_value(entry) for entry in value]
        elif isinstance(value, dict):
            return {
                self.dump_value(x): self.dump_value(y) for x, y in
                value.items()
            }
        else:
            return value

    def dump_property(self, p):
        """Return Property p as a dictionary."""
        pt = {y: x for x, y in self.property_types.items()}
        d = dict(
            type=pt.get(p.type, None), name=p.name, description=p.description,
            value=self.dump_value(p.value)
        )
        if d['type'] is None:
            raise RuntimeError('Invalid type on property %r.' % p)
        return d

    def load_value(self, value):
        """Returns a loaded value."""
        if isinstance(value, ObjectReference):
            return self.objects[value.id]
        elif isinstance(value, list):
            return [self.load_value(entry) for entry in value]
        elif isinstance(value, dict):
            return {
                self.load_value(x): self.load_value(y) for x, y
                in value.items()
            }
        else:
            return value

    def load_property(self, obj, d):
        """Load and return a Property instance bound to an Object instance obj,
        from a dictionary d."""
        try:
            return obj.add_property(
                d['name'], self.property_types.get(d['type']),
                self.load_value(d.get('value', None)),
                description=d.get('description', None)
            )
        except Exception as e:
            raise LoadPropertyError(obj, d) from e

    def dump_method(self, m):
        """Dump a Method m as a dictionary."""
        return dict(name=m.name, code=m.code)

    def load_method(self, obj, d):
        """Load and return a Method instance bound to Object instance obj, from
        a dictionary d."""
        try:
            return obj.add_method(d['code'], name=d.get('name', None))
        except Exception as e:
            raise LoadMethodError(obj, d) from e

    def dump_object(self, obj):
        """Return Object obj as a dictionary."""
        return dict(
            id=obj.id, parents=[parent.id for parent in obj.parents],
            location=obj._location, properties=[
                self.dump_property(p) for p in obj._properties.values()
            ],
            methods=[self.dump_method(m) for m in obj._methods.values()]
        )

    def load_object(self, d):
        """Load and return an Object instance from a dictionary d."""
        try:
            o = self.object_class(self, id=d.get('id', self.max_id))
        except Exception as e:
            raise LoadObjectError(d) from e
        o._location = d.get('location', None)
        self.attach_object(o)
        for data in d.get('methods', []):
            self.load_method(o, data)
        return o

    def dump(self):
        """Generate a dictionary from this database which can be dumped using
        YAML for example."""
        d = dict(objects=[], registered_objects={})
        for name, obj in self.registered_objects.items():
            d['registered_objects'][name] = obj.id
        for obj in sorted(self.objects.values(), key=lambda thing: thing.id):
            d['objects'].append(self.dump_object(obj))
        return d

    def load(self, d):
        """Load objects from a dictionary d."""
        objects = d['objects']
        for data in objects:
            self.load_object(data)
        # All objects are now partially loaded without properties or parents.
        # Let's load the rest.
        for data in objects:
            obj = self.objects[data['id']]
            for datum in data['properties']:
                self.load_property(obj, datum)
            for id in data['parents']:
                obj.add_parent(self.objects[id])
        for name, id in d['registered_objects'].items():
            self.register_object(name, self.objects[id])

    def register_object(self, name, obj):
        """Register an Object instance obj with this database. Once registered,
        it will be available as an attribute."""
        if obj.id is None:
            raise RuntimeError(
                'Cannot register an anonymous object: %r.' % obj
            )
        self.registered_objects[name] = obj

    def unregister_object(self, name):
        """Unregister an Object instance which was previously registered with
        the given name, so it is no longer available as an attribute."""
        del self.registered_objects[name]

    def clear_method_cache(self):
        """Iterate through all objects and clear their method caches. This can
        conserve space, and should probably be run periodically on systems
        where methods change frequently."""
        for obj in self.objects.values():
            obj._method_cache.clear()

    def __getattr__(self, name):
        try:
            return self.registered_objects[name]
        except KeyError:
            return super().__getattribute__(name)



"""Provides the Property class."""

from attr import attrs, attrib

NoneType = type(None)


@attrs
class Property:
    """A property on an Object instance."""

    name = attrib()
    description = attrib()
    type = attrib()
    value = attrib()

    def get(self):
        return self.value

    def set(self, value):
        if not isinstance(value, (NoneType, self.type)):
            raise TypeError(
                'Type mismatch for property %r. Value: %r.' % (self, value)
            )
        self.value = value



"""Test objects."""

from datetime import datetime
from types import MethodType, FunctionType
from pytest import raises
from carehome import Object, Property, Database, Method
from carehome.exc import DuplicateParentError, ParentIsChildError

db = Database()


class CustomObject(Object):
    pass


class CustomProperty(Property):
    pass


class CustomMethod(Method):
    pass


class InvalidType:
    pass


def test_creation():
    o = Object(db)
    assert o.database is db
    assert o.id is None
    assert o._properties == {}
    assert o._methods == {}
    assert o._parents == []
    assert o._children == []
    assert o.parents == []
    assert o.children == []


def test_properties():
    o = Object(db)
    o._properties['hello'] = 'world'
    assert o.hello == 'world'
    with raises(AttributeError):
        print(o.test)


def test_add_parent():
    parent = Object(db)
    child = Object(db)
    child.add_parent(parent)
    assert parent in child.parents
    with raises(DuplicateParentError):
        child.add_parent(parent)
    with raises(DuplicateParentError):
        child.add_parent(child)
    with raises(ParentIsChildError):
        parent.add_parent(child)


def test_properties_inheritance():
    parent = Object(db)
    child = Object(db)
    child.add_parent(parent)
    parent._properties['hello'] = 'world'
    assert child.hello == 'world'
    assert parent.hello == 'world'
    child._properties['hello'] = 'test'
    assert parent.hello == 'world'
    assert child.hello == 'test'


def test_add_property_valid():
    o = Object(db)
    desc = 'Test property'
    value = 'Hello world'
    p = o.add_property('test', str, value, description=desc)
    assert isinstance(p, Property)
    assert p.get() == value
    assert p.type is str
    assert p.description == desc


def test_add_property_invalid():
    o = Object(db)
    with raises(TypeError):
        o.add_property('test1', str, datetime.utcnow())
    with raises(TypeError):
        o.add_property('test2', InvalidType, 'Hello world.')


def test_add_property_null():
    o = Object(db)
    p = o.add_property('location', Object, None)
    assert p.value is None
    assert p.name == 'location'
    assert p.type is Object


def test_add_property_duplicate_name():
    o = Object(db)
    name = 'test'
    value = 'Hello world.'
    o.add_property(name, str, value)
    with raises(NameError):
        o.add_property(name, type, value)


def test_add_property_custom():
    d = Database(property_class=CustomProperty)
    o = d.create_object()
    p = o.add_property('test', str, 'Hopefully works.')
    assert isinstance(p, CustomProperty)


def test_property_get():
    parent = Object(db)
    value = datetime.utcnow()
    parent.add_property('date', datetime, value)
    assert parent.date is value
    child = Object(db)
    child.add_parent(parent)
    assert child.date is value
    string = '02/10/2018'
    child.add_property('date', str, string)
    assert parent.date is value
    assert child.date == string


def test_add_method():
    o = db.create_object()
    o.add_method('def test(self):\n    return self')
    assert o.test() is o


def test_add_method_anonymous():
    o = Object(db)
    with raises(RuntimeError):
        o.add_method(
            'fails', 'return "This will fail because the object is anonymous."'
        )


def test_add_method_custom():
    d = Database(method_class=CustomMethod)
    o = d.create_object()
    m = o.add_method('def test(self):\n    pass\n')
    assert isinstance(m, CustomMethod)


def test_remove_method():
    o = db.create_object()
    o.add_method('def test(self):\n    return')
    assert callable(o.test)
    o.remove_method('test')
    assert not hasattr(o, 'test')


def test_set_id():
    o = db.create_object()
    with raises(RuntimeError):
        o.id = 2


def test_set___initialised__():
    o = db.create_object()
    with raises(RuntimeError):
        o.__initialised__ = None


def test_parents():
    parent_1 = db.create_object()
    parent_2 = db.create_object()
    grandparent_1 = db.create_object()
    o = db.create_object()
    for parent in (parent_1, parent_2):
        o.add_parent(parent)
    assert o.parents == [parent_1, parent_2]
    parent_1.add_parent(grandparent_1)
    assert o.parents == [parent_1, parent_2]


def test_children():
    parent = db.create_object()
    child_1 = db.create_object()
    child_2 = db.create_object()
    for child in (child_1, child_2):
        child.add_parent(parent)
    assert parent.children == [child_1, child_2]
    grandchild_1 = db.create_object()
    grandchild_1.add_parent(child_1)
    assert parent.children == [child_1, child_2]


def test_ancestors():
    grandparent_1 = db.create_object()
    grandparent_2 = db.create_object()
    grandparent_3 = db.create_object()
    grandparent_4 = db.create_object()
    parent_1 = db.create_object()
    parent_2 = db.create_object()
    o = db.create_object()
    assert list(o.ancestors()) == []
    for parent in (parent_1, parent_2):
        o.add_parent(parent)
    assert list(o.ancestors()) == [parent_1, parent_2]
    for grandparent in (grandparent_1, grandparent_2):
        parent_1.add_parent(grandparent)
    for grandparent in (grandparent_3, grandparent_4):
        parent_2.add_parent(grandparent)
    ancestors = list(o.ancestors())
    assert ancestors == [
        parent_1, grandparent_1, grandparent_2, parent_2, grandparent_3,
        grandparent_4
    ]


def test_descendants():
    grandparent = db.create_object()
    parent_1 = db.create_object()
    parent_2 = db.create_object()
    child_1 = db.create_object()
    child_2 = db.create_object()
    child_3 = db.create_object()
    child_4 = db.create_object()
    assert list(grandparent.descendants()) == []
    for parent in (parent_1, parent_2):
        parent.add_parent(grandparent)
    assert list(grandparent.descendants()) == [parent_1, parent_2]
    for child in (child_1, child_2):
        child.add_parent(parent_1)
    for child in (child_3, child_4):
        child.add_parent(parent_2)
    descendants = list(grandparent.descendants())
    assert descendants == [
        parent_1, child_1, child_2, parent_2, child_3, child_4
    ]


def test_get_method():
    parent = db.create_object()
    m = parent.add_method('def this(self):\n    return self')
    assert isinstance(m.func, FunctionType)
    assert isinstance(parent.this, MethodType)
    assert parent.this.__self__ is parent
    assert parent.this() is parent
    child = db.create_object()
    child.add_parent(parent)
    assert isinstance(child.this, MethodType)
    assert child.this is not parent.this
    assert child.this.__self__ is child
    assert child.this() is child


def test_method_cache():
    o = db.create_object()
    m = o.add_method('def test(self):\n    return 1')
    assert o.test() == 1
    assert o._method_cache == {id(m.func): o.test}
    o.remove_method('test')
    m = o.add_method('def test(self):\n    return 2')
    assert o.test() == 2
    assert o._method_cache[id(m.func)] is o.test


def test_location():
    loc = db.create_object()
    assert loc.location is None
    assert not loc.contents
    aside = db.create_object(loc)
    assert aside.location is None
    assert not aside.contents
    # Creating a new object shouldn't affect any previously-loaded objects.
    assert loc.location is None
    assert not loc.contents
    thing = db.create_object()
    thing.location = loc
    assert loc.contents == [thing]
    assert thing.location is loc
    assert thing._location == loc.id
    thing.location = None
    assert thing.location is None
    assert not loc.contents


def test_location_custom_object():
    db = Database(object_class=CustomObject)
    loc = db.create_object()
    assert isinstance(loc, CustomObject)
    obj = db.create_object()
    assert isinstance(obj, CustomObject)
    obj.location = loc
    assert obj.location is loc
    assert obj._location == loc.id
    assert loc.contents == [obj]


def test_find_property():
    db = Database()
    parent = db.create_object()
    name = 'test'
    value = 'Test value'
    description = 'Test property for testing purposes.'
    p = parent.add_property(name, str, value, description=description)
    o = db.create_object(parent)
    assert o.find_property(name) is p
    assert o.find_property('fake') is None


def test_property_dynamic_add():
    db = Database()
    parent = db.create_object()
    name = 'test'
    value = 'Testing'
    other_value = 'This is a different value.'
    description = 'Test property.'
    parent.add_property(name, str, value, description=description)
    o = db.create_object(parent)
    o.test = other_value
    assert o.test == other_value
    p = o._properties[name]
    assert p.name == name
    assert p.value == other_value
    assert p.description == description
    o.other = value
    p = o._properties['other']
    assert p.name == 'other'
    assert p.value == value
    assert p.description == 'Added by __setattr__.'
    assert p.type is str



"""Test the Method class."""

import os
import os.path
import re
from inspect import isclass, isfunction
from types import FunctionType
import flake8
from pytest import raises
from carehome import Method, Database, methods
from carehome.exc import Flake8NotFound

db = Database()
inserted_global = object()

code = """
class Thing:
    pass


def f1():
    return 1


def f2():
    return 2


def f():
    return (Thing, f1, f2)
"""


def test_init():
    name = 'test'
    description = 'This is a test.'
    m = Method(
        db, 'def %s(self):\n    """%s"""\n    print("Hello world.")' % (
            name, description
        )
    )
    assert isinstance(m.func, FunctionType)
    assert m.name == name
    f = m.func
    assert f.__name__ == name
    assert f.__doc__ == description


def test_get_filename():
    m = Method(db, 'def test_filename():\n    print()', name='test_filename')
    assert os.path.isfile(m.get_filename())
    with open(m.get_filename(), 'r') as f:
        assert f.read() == m.code


def test_run():
    m = Method(
        db, 'def test():\n    """Test function."""\n    return 12345'
    )
    assert m.func() == 12345


def test_args():
    m = Method(
        db, 'def test(a, b=5):\n    """Test function."""\n    return (a, b)'
    )
    assert m.func(1) == (1, 5)
    assert m.func(4, b=10) == (4, 10)


def test_imports():
    m = Method(db, 'import re\ndef test():\n    """Test re."""\n    return re')
    assert m.func() is re


def test_method_globals():
    db = Database(method_globals=dict(g=inserted_global))
    o = db.create_object()
    o.add_method('def get_g(self):\n    return g')
    assert o.get_g() is inserted_global


def test_created():
    m = Method(db, code, name='f')
    Thing, f1, f2 = m.func()
    # Make sure the right stuff has been returned.
    assert isclass(Thing)
    assert isfunction(f1)
    assert isfunction(f2)
    assert f1() == 1
    assert f2() == 2
    assert m.created['Thing'] is Thing
    assert m.created['f1'] is f1
    assert m.created['f2'] is f2


def test_method_guess_name():
    m = Method(db, 'def f():\n    return 1234')
    assert m.name == 'f'
    assert m.func() == 1234


def test_no_func():
    with raises(RuntimeError):
        Method(db, 'import sys')


def test_flake8():
    assert methods.flake8 is flake8
    m = Method(db, 'def f():\n    pass\n')
    methods.flake8 = None
    with raises(Flake8NotFound):
        m.validate_code()
    methods.flake8 = flake8
    assert m.validate_code() is None
    m = Method(db, 'def f():\n    return something\n', name='f')
    with raises(NameError):
        m.func()
    expected = "stdin:2:12: F821 undefined name 'something'%s" % os.linesep
    res = m.validate_code()
    assert res == expected
    db.method_globals['pretend'] = 1234
    m = Method(db, 'def f():\n    return pretend\n', name='f')
    assert m.validate_code() is None



"""Provides the Object class."""

from types import MethodType
from attr import attrs, attrib, Factory
from .exc import DuplicateParentError, ParentIsChildError, NoSuchEventError

NoneType = type(None)


@attrs
class Object:
    """An object with multiple parents and multiple children."""

    database = attrib()
    _parents = attrib(default=Factory(list))
    _children = attrib(default=Factory(list))
    _methods = attrib(default=Factory(dict))
    _properties = attrib(default=Factory(dict))
    id = attrib(default=Factory(type(None)))
    _method_cache = attrib(default=Factory(dict), init=False, repr=False)
    _location = attrib(default=Factory(NoneType))

    def __attrs_post_init__(self):
        self.__initialised__ = True

    def __setattr__(self, name, value):
        initialised = '__initialised__' in self.__dict__
        reserved_names = ('__initialised__', 'id')
        if initialised and name in reserved_names:
                raise RuntimeError(
                    'You cannot set this attribute after initialisation.'
                )
        if not initialised or (
            name in self.__dict__ or name in dir(self.database.object_class)
        ):
            return super().__setattr__(name, value)
        elif name in self._properties:
            self._properties[name].set(value)
        else:
            prop = self.find_property(name)
            if prop is None:
                t = type(value)
                description = 'Added by __setattr__.'
            else:
                t = prop.type
                description = prop.description
            self.add_property(name, t, value, description=description)

    @property
    def parents(self):
        return self._parents.copy()

    @property
    def children(self):
        return self._children.copy()

    @property
    def methods(self):
        return self._methods.keys()

    @property
    def properties(self):
        return self._properties.keys()

    def descendants(self):
        """Return all descendants of this object."""
        for child in self._children:
            yield child
            for descendant in child.descendants():
                yield descendant

    def ancestors(self):
        """Return all the ancestors of this object."""
        for parent in self._parents:
            yield parent
            for ancestor in parent.ancestors():
                yield ancestor

    @property
    def location(self):
        if self._location is not None:
            return self.database.objects[self._location]

    @location.setter
    def location(self, obj):
        """Set the location of this object to a destination where. Note that
        where must either be an Object instance, or None, which means
        nowhere."""
        if self._location is not None:
            self.location.try_event('on_exit', self.location, self)
        if obj is None:
            value = None
        else:
            obj.try_event('on_enter', obj, self)
            value = obj.id
        self.__dict__['_location'] = value

    @property
    def contents(self):
        objects = self.database.objects.values()
        return [x for x in objects if x.location is self]

    def add_parent(self, obj):
        """Add a parent to this object."""
        assert isinstance(obj, type(self))
        if obj in self.descendants():
            raise ParentIsChildError(self, obj)
        if obj in self.ancestors() or obj is self:
            raise DuplicateParentError(self, obj)
        self.try_event('on_add_parent', self, obj)
        obj.try_event('on_add_child', obj, self)
        self._parents.append(obj)
        obj._children.append(self)

    def remove_parent(self, obj):
        """Remove a parent from this object."""
        self.try_event('on_remove_parent', self, obj)
        obj.try_event('on_remove_child', obj, self)
        self._parents.remove(obj)
        obj._children.remove(self)

    def method_or_property(self, attribute):
        """Get a method or property with the given name."""
        d = {}
        for dictionary in (self._properties, self._methods):
            d.update(dictionary)
        for name, value in d.items():
            if name == attribute:
                return value
        raise AttributeError(attribute)

    def __getattr__(self, name, *args, **kwargs):
        """Find a property or method matching the given name."""
        try:
            value = self.method_or_property(name)
        except AttributeError:
            for ancestor in self.ancestors():
                try:
                    value = ancestor.method_or_property(name)
                    break
                except AttributeError:
                    pass
            else:
                return super().__getattribute__(name, *args, **kwargs)
        if isinstance(value, self.database.property_class):
            return value.get()
        elif isinstance(value, self.database.method_class):
            num = id(value.func)
            if num not in self._method_cache:
                self._method_cache[num] = MethodType(value.func, self)
            return self._method_cache[num]
        else:
            return value

    def add_property(self, name, type, value, description=None):
        """Add a property to this Object."""
        if name in self._properties:
            raise NameError('Duplicate property name: %r.' % name)
        for cls in self.database.property_types.values():
            if cls is type:
                break
        else:
            raise TypeError(
                'Invalid property type for %r.%s (value=%r): %r.' % (
                    self, name, value, type
                )
            )
        if not isinstance(value, (NoneType, type)):
            raise TypeError('Value %r is not of type %r.' % (value, type))
        p = self.database.property_class(name, description, type, value)
        self.try_event('on_add_property', self, p)
        self._properties[name] = p
        return p

    def remove_property(self, name):
        """Remove a property from this object."""
        self.try_event('on_remove_property', self, name)
        del self._properties[name]

    def find_property(self, name):
        """Fnd a property with the given name and return it."""
        objects = [self]
        objects.extend(self.ancestors())
        for obj in objects:
            if name in obj.properties:
                return obj._properties[name]

    def add_method(self, *args, **kwargs):
        """Add a method to this object. All arguments are passed to the Method
        constructor. The first method argument must be self or similar, so this
        object can be available from within the function itself. Methods can
        not be added to anonymous objects (those with no IDs)."""
        if self.id is None:
            raise RuntimeError('Methods cannot be added to anonymous objects.')
        m = self.database.method_class(self.database, *args, **kwargs)
        self._methods[m.name] = m
        return m

    def remove_method(self, name):
        """Remove a method from this object."""
        del self._methods[name]

    def do_event(self, name, *args, **kwargs):
        """Call the named event with the given args and kwargs."""
        if callable(getattr(self, name, None)):
            return getattr(self, name)(*args, **kwargs)
        raise NoSuchEventError(self, name, args, kwargs)

    def try_event(self, name, *args, **kwargs):
        """Tries to run the given event. The return value is either None if the
        event is not present, or the return value of the vent method."""
        try:
            return self.do_event(name, *args, **kwargs)
        except NoSuchEventError:
            pass



"""Test Database objects."""

import re
from datetime import datetime
from types import FunctionType
from pytest import raises
from carehome import Database, Object, Property, Method, ObjectReference
from carehome.exc import (
    LoadPropertyError, LoadMethodError, LoadObjectError, ObjectRegisteredError,
    HasChildrenError, HasContentsError, IsValueError
)


class CustomObject(Object):
    pass


def test_create():
    db = Database()
    assert db.objects == {}
    assert db.max_id == 0
    assert db.method_globals == {'database': db, 'objects': db.objects}
    assert db.methods_dir == 'methods'
    assert db.object_class is Object
    assert db.property_class is Property
    assert db.method_class is Method


def test_create_object():
    db = Database()
    o1 = db.create_object()
    assert isinstance(o1, Object)
    assert o1.database is db
    assert o1.id == 0
    assert db.max_id == 1
    assert db.objects == {0: o1}
    o2 = db.create_object()
    assert isinstance(o2, Object)
    assert o2.database is db
    assert o2.id == 1
    assert db.max_id == 2
    assert db.objects == {0: o1, 1: o2}


def test_destroy_object():
    db = Database()
    o = db.create_object()
    db.destroy_object(o)
    assert db.objects == {}
    o1 = db.create_object()
    o2 = db.create_object()
    db.destroy_object(o2)
    assert o1.id == 1
    assert db.objects == {1: o1}
    assert db.max_id == 3


def test_destroy_object_with_parents():
    d = Database()
    g = d.create_object()
    p = d.create_object(g)
    # I know we've done these inheritence tests elsewhere, I just feel more
    # secure knowing they're in two separate places.
    assert g.children == [p]
    assert p.parents == [g]
    c = d.create_object(p)
    assert c.parents == [p]
    assert p.children == [c]
    d.destroy_object(c)
    assert not p.children
    assert g.children == [p]


def test_destroy_object_with_children():
    d = Database()
    parent = d.create_object()
    d.create_object(parent)
    d.create_object(parent)
    with raises(HasChildrenError) as exc:
        d.destroy_object(parent)
    assert exc.value.args[0]is parent
    for obj in parent.children:
        obj.remove_parent(parent)
    d.destroy_object(parent)
    assert parent.id not in d.objects


def test_destroy_object_with_contents():
    d = Database()
    room = d.create_object()
    obj_1 = d.create_object()
    obj_1.location = room
    obj_2 = d.create_object()
    obj_2.location = room
    with raises(HasContentsError) as exc:
        d.destroy_object(room)
    assert exc.value.args[0] is room
    for obj in room.contents:
        obj.location = None
    d.destroy_object(room)
    assert room.id not in d.objects


def test_destroy_stored_object():
    d = Database()
    o1 = d.create_object()
    o2 = d.create_object()
    p = o1.add_property('prop', d.object_class, o2)

    def inner():
        """Actually perform the testing."""
        with raises(IsValueError) as exc:
            d.destroy_object(o2)
        obj, prop = exc.value.args
        assert obj is o1
        assert prop is p

    inner()
    p.type = list
    p.set([1, 2, o2])
    inner()
    p.type = dict
    p.set(dict(obj=o2))
    inner()
    p.set(dict(objects=[o1, o2]))
    inner()
    p.value = None
    d.destroy_object(o2)
    assert o2.id not in d.objects


def test_destroy_object_registered():
    d = Database()
    first = d.create_object()
    d.register_object('first', first)
    with raises(ObjectRegisteredError):
        d.destroy_object(first)
    assert d.first is first
    assert d.objects == {first.id: first}


def test_dump_property():
    d = Database()
    name = 'test'
    desc = 'Test property.'
    type = str
    value = 'Test string.'
    p = Property(name, desc, type, value)
    actual = d.dump_property(p)
    expected = dict(name=name, description=desc, value=value, type='str')
    assert expected == actual
    p.type = Exception
    with raises(RuntimeError):
        d.dump_property(p)


def test_load_property():
    d = Database()
    o = d.create_object()
    name = 'test'
    desc = 'Test date.'
    value = datetime.utcnow()
    p = d.load_property(
        o, dict(name=name, type='datetime', value=value, description=desc)
    )
    assert p.value == value
    assert p.type is datetime
    assert p.description == desc
    assert o._properties[p.name] is p


def test_load_property_malformed():
    d = Database()
    o = d.create_object()
    with raises(LoadPropertyError) as err:
        d.load_property(o, d)
    assert err.value.args[0] is o
    assert err.value.args[1] is d
    assert isinstance(err.value.__cause__, TypeError)


def test_dump_method():
    d = Database()
    name = 'test'
    code = 'def %s(self):\n    return 1234' % name
    m = Method(d, code)
    actual = d.dump_method(m)
    expected = dict(name=name, code=code)
    assert actual == expected


def test_load_method():
    d = Database()
    o = d.create_object()
    name = 'test'
    code = 'import re\ndef %s(self, a, b):\n    return (a, b, re)' % name
    m = d.load_method(o, dict(name=name, code=code))
    assert isinstance(m, Method)
    assert o._methods[name] is m
    assert m.database is d
    assert m.name == name
    assert isinstance(m.func, FunctionType)
    assert m.code == code
    assert m.func(o, 1, 2) == (1, 2, re)


def load_method_malformed():
    d = Database()
    o = d.create_object()
    with raises(LoadMethodError) as err:
        d.load_metod(o, d)
    assert err.value.args[0] is o
    assert err.value.args[1] is d
    assert isinstance(err.value.__cause__, TypeError)


def test_dump_object():
    d = Database()
    parent_1 = d.create_object()
    parent_2 = d.create_object()
    assert d.dump_object(parent_1) == dict(
        location=None, id=parent_1.id, parents=[], properties=[], methods=[]
    )
    o = d.create_object()
    for parent in (parent_1, parent_2):
        o.add_parent(parent)
    actual = d.dump_object(o)
    expected = dict(
        location=None, id=o.id, properties=[], methods=[],
        parents=[parent_1.id, parent_2.id]
    )
    assert actual == expected
    parent_2.location = parent_1
    res = d.dump_object(parent_2)
    assert 'location' not in res['properties']
    assert res['location'] == parent_1.id


def test_load_object():
    d = Database()
    id = 18
    p = Property('property', 'Test property.', bool, False)
    m = Method(d, 'import re\ndef method(self, a, b):\n    return (a, b, re)')
    data = dict(
        id=id, methods=[d.dump_method(m)], properties=[d.dump_property(p)],
        parents=[]
    )
    o = d.load_object(data)
    assert d.objects[id] is o
    assert d.max_id == (id + 1)
    assert len(o._methods) == 1
    m.func = None
    for method in (m, o._methods[m.name]):
        method.created.clear()
    o._methods[m.name].func = None  # Otherwise they'll never match.
    assert o._methods[m.name] == m
    assert not o._properties
    assert not o._parents
    assert not o._children
    data = dict(id=1, location=2)
    o = d.load_object(data)
    assert o._location == 2


def test_load_object_malformed():
    d = Database()
    with raises(LoadObjectError) as err:
        d.load_object(d)
    assert err.value.args[0] is d
    assert isinstance(err.value.__cause__, AttributeError)


def test_dump():
    d = Database()
    o1 = d.create_object()
    o2 = d.create_object()
    name = 'test'
    d.register_object(name, o1)
    data = d.dump()
    assert data['registered_objects'] == {name: o1.id}
    assert len(data['objects']) == 2
    assert data['objects'][0]['id'] == o1.id
    assert data['objects'][1]['id'] == o2.id


def test_load():
    d = Database()
    grandparent_1 = d.create_object()
    grandparent_2 = d.create_object()
    parent = d.create_object()
    for grandparent in (grandparent_1, grandparent_2):
        parent.add_parent(grandparent)
    child = d.create_object()
    child.add_parent(parent)
    for name, obj in (
        ('grandparent_1', grandparent_1), ('grandparent_2', grandparent_2),
        ('parent', parent), ('child', child)
    ):
        d.register_object(name, obj)
    data = d.dump()
    new = Database()
    new.load(data)
    assert d.child.id == child.id
    assert d.parent.id == parent.id
    assert d.child._parents == [d.parent]
    assert d.grandparent_1.id == grandparent_1.id
    assert d.grandparent_2.id == grandparent_2.id
    assert d.parent._parents == [d.grandparent_1, d.grandparent_2]


def test_register_object():
    d = Database()
    o = d.create_object()
    name = 'test'
    d.register_object(name, o)
    assert d.registered_objects == {name: o}
    assert d.test is o
    o = Object(d)
    with raises(RuntimeError):
        d.register_object(name, o)


def test_unregister_object():
    d = Database()
    o = d.create_object()
    name = 'test'
    d.register_object(name, o)
    d.unregister_object(name)
    assert not d.registered_objects
    with raises(AttributeError):
        print(d.test)
    with raises(KeyError):
        d.unregister_object(name)


def test_dump_objectreference():
    d = Database()
    f1 = d.create_object()
    f2 = d.create_object()
    o = d.create_object()
    p = o.add_property('friends', list, [f1, f2])
    assert p.value == [f1, f2]
    data = d.dump_property(p)
    for entry in data['value']:
        assert isinstance(entry, ObjectReference), 'Invalid entry: %r.' % entry
    assert data['value'] == [ObjectReference(f1.id), ObjectReference(f2.id)]


def test_dump_value():
    d = Database()
    o1 = d.create_object()
    o2 = d.create_object()
    data = dict(names=['hello', 'world'], age=25, objects=[o1, o2])
    res = d.dump_value(data)
    assert res == dict(
        names=['hello', 'world'], age=25, objects=[
            ObjectReference(o1.id), ObjectReference(o2.id)
        ]
    )


def test_load_value():
    d = Database()
    o1 = d.create_object()
    o2 = d.create_object()
    data = dict(
        names=['hello', 'world'], objects=[
            ObjectReference(o1.id), ObjectReference(o2.id)
        ], age=29
    )
    value = d.load_value(data)
    assert value == dict(
        names=['hello', 'world'], objects=[o1, o2], age=29
    )


def test_custom_object():
    d = Database(object_class=CustomObject)
    o = d.create_object()
    assert isinstance(o, d.object_class)
    o = d.load_object(dict(id=1, methods=[]))
    assert isinstance(o, CustomObject)


def test_dump_value_object_class():
    d = Database(object_class=CustomObject)
    o = d.create_object()
    value = d.dump_value(o)
    assert isinstance(value, ObjectReference)


def test_load_value_object_class():
    d = Database(object_class=CustomObject)
    o = d.create_object()
    assert isinstance(o, CustomObject)
    value = d.load_value(ObjectReference(o.id))
    assert value is o


def test_clear_func_cache():
    d = Database()
    parent = d.create_object()
    parent.add_method('def test1(self):\n    return 1')
    child = d.create_object()
    child.add_method('def test2(self):\n    return 2')
    child.add_parent(parent)
    parent.test1()
    assert len(parent._method_cache) == 1
    child.test1()
    child.test2()
    assert len(child._method_cache) == 2
    d.clear_method_cache()
    assert not parent._method_cache
    assert not child._method_cache


def test_create_object_with_parents():
    d = Database()
    parent_1 = d.create_object()
    parent_2 = d.create_object()
    child = d.create_object(parent_1, parent_2)
    assert not parent_1.parents
    assert not parent_2.parents
    assert child.parents == [parent_1, parent_2]


def test_property_types():
    d = Database()
    assert d.property_types['obj'] is Object
    d = Database(object_class=CustomObject)
    assert d.property_types['obj'] is CustomObject
    o = d.create_object()
    o.this = o
    assert o._properties['this'].type is d.object_class


def test_dump_property_custom_object():
    d = Database(object_class=CustomObject)
    o = d.create_object()
    p = o.add_property('this', d.object_class, o)
    value = d.dump_property(p)
    assert value['type'] == 'obj'



"""Provides exception classes."""


class CarehomeError(Exception):
    """All other errors should inherit from this class."""


class InheritanceError(CarehomeError):
    """An error in the inheritance hierarchy."""


class DuplicateParentError(InheritanceError):
    """This object already has this parent."""


class DuplicateChildError(InheritanceError):
    """This object already has this child."""


class ParentIsChildError(InheritanceError):
    """This parent is already a child of this object."""


class DatabaseError(CarehomeError):
    """There is a problem in the database."""


class NoSuchEventError(CarehomeError):
    """No such event."""


class Flake8NotFound(CarehomeError):
    """Flake8 not found."""


class LoadError(CarehomeError):
    """Error loading something."""


class LoadPropertyError(LoadError):
    """Error loading a property."""


class LoadMethodError(LoadError):
    """Error loading a method."""


class LoadObjectError(LoadError):
    """Error loading an object."""


class DestroyError(CarehomeError):
    """Error destroying an object."""


class ObjectRegisteredError(DestroyError):
    """Object is registered."""


class HasChildrenError(DestroyError):
    """This object has children."""


class HasContentsError(DestroyError):
    """Object has contents."""


class IsValueError(DestroyError):
    """This object is stored in the property of another object."""



"""The carehome module. Allows you to build MOO-like object-oriented objects in
Python."""

from .objects import Object
from .properties import Property
from .property_types import property_types
from .methods import Method
from .databases import Database, ObjectReference

__all__ = ['property_types']

for thing in (
    Object, Property, Method, Database, ObjectReference
):
    __all__.append(thing.__name__)



"""Provides the Method class."""

import os
import os.path
from inspect import isfunction
from subprocess import Popen, PIPE
from attr import attrs, attrib, Factory
try:
    import flake8
except ImportError:
    flake8 = None
from .exc import Flake8NotFound

NoneType = type(None)


@attrs
class Method:
    """An Object method."""

    database = attrib()
    code = attrib()
    name = attrib(default=Factory(NoneType))
    func = attrib(default=Factory(NoneType), init=False)
    created = attrib(default=Factory(dict), init=False)

    def __attrs_post_init__(self):
        g = globals().copy()
        g.update(**self.database.method_globals)
        old_names = set(g.keys())
        n = self.get_filename()
        with open(n, 'w') as f:
            f.write(self.code)
        source = compile(self.code, n, 'exec')
        eval(source, g)
        new_names = set(g.keys())
        for name in new_names.difference(old_names):
            f = g[name]
            self.created[name] = f
            if self.name is None and isfunction(f):
                self.name = name
        if self.name is None:
            raise RuntimeError('No function found.')
        self.func = self.created[self.name]

    def get_filename(self):
        """Get a unique filename for this method."""
        return os.path.join(
            self.database.methods_dir, '%s-%d.method' % (self.name, id(self))
        )

    def validate_code(self):
        """This method by default uses Flake8 to check your code. It should
        return either None to indicate no errors, or a string containing any
        problems found."""
        if flake8 is None:
            raise Flake8NotFound()
        builtins = ','.join(self.database.method_globals.keys())
        p = Popen(
            ('flake8', '--builtins=%s' % builtins, '-'), stdin=PIPE,
            stdout=PIPE, stderr=PIPE
        )
        stdout, stderr = p.communicate(self.code.encode())
        if stdout:
            return stdout.decode()
        elif stderr:
            return stderr.decode()
